//
//
// @author Michael Liljenstam, ISTS, Dartmouth College
//
// $Id: CodeRedWormEpidemicInitializer.java,v 1.1 2003/03/01 21:04:07 mili Exp $

package SSF.App.Worm;

import java.util.*;


/**
 *  Initialize worm epidemic using empirical distributions based on 
 *  routing table data and published data on Code Red II infections in 
 *  August 2001.
 *  <p>
 *  Caveat: <br>
 *  Assigns IP space according to empirical distribution over ASes, 
 *  but doesn't take (probable) correlation with topology into account.
 *  Distribution of susceptibles follows from IP space. Thus, it will 
 *  possibly also lack this structure.
 */
public class CodeRedWormEpidemicInitializer extends WormEpidemicInitializer {

    /** 
     *  Perform the initialization.
     */
    public void doInit() {
	// initalize empirical distributions
	IPSpaceDistr.init(true, // use theoretical approximation
			  MacroscopicModel.ref.infRNG); // use infection RNG
	CR2SuscFracDistr.init(false, // use empirical distribution
			      MacroscopicModel.ref.infRNG);

	// set local model vars
	ASGraph asGraph = MacroscopicModel.ref.net;
	int numASs = asGraph.numASs;

	float Beta = MacroscopicModel.ref.wormModel.Beta;
	float sTot = MacroscopicModel.ref.wormModel.sTot;
	float iTot = MacroscopicModel.ref.wormModel.iTot;
	float rTot = MacroscopicModel.ref.wormModel.rTot;

	// sample IP space sizes
	long ipSpaces[] = new long[numASs];
	long totIpSpace = 0;
	for (int j = 0; j < numASs; j++) {
	    ipSpaces[j] = IPSpaceDistr.nextLong();
	    totIpSpace += ipSpaces[j];
	} // for

	// init infection parameters
	float[] BetaJ  = new float[numASs];
	MacroscopicModel.ref.wormModel.BetaJ = BetaJ;
	for (int j = 0; j < numASs; j++) {
	    // set the cross-strata infection parameters according to the 
	    // fraction of IP space announced by each AS
	    BetaJ[j] = 
		(float) (Beta * (ipSpaces[j] / (double)totIpSpace) / numASs);
	} // for

	// init susceptibles
	LinkedList leaves = new LinkedList();
	if (numASs > 1) {
	    // stratified model
	    for (int j = 0; j < numASs; j++) {
		if (asGraph.ases[j].adjacencyList.size() == 1)
		    leaves.add(asGraph.ases[j]);
	    }
	} else {
	    // homogeneous model
	    leaves.add(asGraph.ases[0]);
	}
	int numLeaves = leaves.size();
	for (int i = 0; i < numLeaves; i++) {
	    AS as = (AS) leaves.removeFirst();
	    as.wormModel.s = sTot / numLeaves; // Uniform distr over leaves
	}

	// init initially infected	    
	asGraph.ases[0].wormModel.i = iTot; // all initally infected in AS 0

	if (MacroscopicModel.ref.wormModel.DEBUG) {
	    System.out.println("Init worm epidemic model: s=" 
			       + sTot + ", i=" + iTot + ", r=" + rTot
			       + ", Beta=" + Beta);
	}
    } // doInit()
}
